/* This file (c) Steven Boswell, 2005
Jerome Cornet <jcornet@free.fr> added vertical flip and command line options
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#define HAVE_STDINT_H
#include "mjpegtools/mjpeg_types.h"
#include "mjpegtools/yuv4mpeg.h"
#include "mjpegtools/mjpeg_logging.h"

int
main (int argc, char *argv[])
{
  int errno = 0;
  struct
  {
    int w, h;			/* width/height of intensity frame */
    int Cw, Ch;			/* width/height of color frame */
    int ss_h, ss_v;		/* ratio between intensity/color sizes */
    uint8_t *in[3];		/* frame data read from stdin */
    uint8_t *out[3];		/* frame data written to stdout */
  } frame;
  y4m_frame_info_t frameinfo;
  y4m_stream_info_t streaminfo;
  int chroma_mode;
  int flip_horizontally = 0;
  int flip_vertically = 0;
  char c;
  long frame_number = 0;
  int verbose = 0;

  while ((c = getopt (argc, argv, "hvHVxy")) != -1)
    {
      switch (c)
	{
	case 'h':
	  {
	    mjpeg_log (LOG_INFO,
		       " -------------------------------------------------");
	    mjpeg_log (LOG_INFO,
		       "            MJPEG Tools Image flipper             ");
	    mjpeg_log (LOG_INFO,
		       " -------------------------------------------------");

	    mjpeg_log (LOG_INFO, " Usage ");
	    mjpeg_log (LOG_INFO,
		       " yuvflip  -[HV] < stream_in.yuv > stream_out.yuv");
	    mjpeg_log (LOG_INFO, " -h: this message");
	    mjpeg_log (LOG_INFO, " -v: verbose output");
	    mjpeg_log (LOG_INFO, " -H or -x: flips the image horizontally");
	    mjpeg_log (LOG_INFO, " -V or -y: flips the image vertically");
	    exit (0);
	    break;
	  }
	case 'x':
	case 'H':
	  {
	    flip_horizontally = 1;
	    break;
	  }
	case 'y':
	case 'V':
	  {
	    flip_vertically = 1;
	    break;
	  }
	case 'v':
	  {
	    verbose = 1;
	    break;
	  }
	default:
	  mjpeg_log (LOG_WARN, "Ignoring parameter: %c", c);
	}
    }
  if ((!flip_horizontally) && (!flip_vertically))
    {
#ifndef OLD_BEHAVIOUR
      mjpeg_log (LOG_ERROR, "Doing nothing, please specify -H -V or both");
      exit (1);
#else
      flip_horizontally = 1;
#endif
    }
  if (verbose)
    {

      mjpeg_log (LOG_INFO,
		 " -------------------------------------------------");
      mjpeg_log (LOG_INFO,
		 "            MJPEG Tools Image flipper             ");
      mjpeg_log (LOG_INFO,
		 " -------------------------------------------------");
    }

  /* initialize stream-information */
  y4m_accept_extensions (1);
  y4m_init_stream_info (&streaminfo);
  y4m_init_frame_info (&frameinfo);

  /* open input stream */
  if ((errno = y4m_read_stream_header (0, &streaminfo)) != Y4M_OK)
    mjpeg_error_exit1 ("Couldn't read YUV4MPEG header: %s!",
		       y4m_strerr (errno));
  if (y4m_si_get_plane_count (&streaminfo) != 3)
    mjpeg_error_exit1 ("Only 3-plane formats supported.");
  frame.w = y4m_si_get_width (&streaminfo);
  frame.h = y4m_si_get_height (&streaminfo);
  chroma_mode = y4m_si_get_chroma (&streaminfo);
  frame.ss_h = y4m_chroma_ss_x_ratio (chroma_mode).d;
  frame.ss_v = y4m_chroma_ss_y_ratio (chroma_mode).d;
  frame.Cw = frame.w / frame.ss_h;
  frame.Ch = frame.h / frame.ss_v;

  if (verbose)
    mjpeg_log (LOG_INFO, "Dimensions: w:%d h:%d Cw:%d Ch:%d", frame.w,
	       frame.h, frame.Cw, frame.Ch);

  /* allocate memory for frames */
  {
    int luma_buffsize;
    int chroma_buffsize;

    luma_buffsize = frame.w * frame.h;
    chroma_buffsize = frame.Cw * frame.Ch;

    frame.in[0] = (uint8_t *) malloc (luma_buffsize);
    frame.in[1] = (uint8_t *) malloc (chroma_buffsize);
    frame.in[2] = (uint8_t *) malloc (chroma_buffsize);
    frame.out[0] = (uint8_t *) malloc (luma_buffsize);
    frame.out[1] = (uint8_t *) malloc (chroma_buffsize);
    frame.out[2] = (uint8_t *) malloc (chroma_buffsize);

    if (frame.in[0] == NULL
	|| frame.in[1] == NULL
	|| frame.in[2] == NULL
	|| frame.out[0] == NULL
	|| frame.out[1] == NULL || frame.out[2] == NULL)
      mjpeg_error_exit1 ("Out of memory " "when allocating frame buffers");
  }

  /* write the outstream header */
  y4m_write_stream_header (1, &streaminfo);

  /* read every single frame until the end of the input stream */
  while (errno = y4m_read_frame (0, &streaminfo, &frameinfo,
				 frame.in), errno == Y4M_OK)
    {
      int x, y, i, ri;

      if (flip_horizontally)
	{

	  /* Flip the frame horizontally. */
	  for (y = 0; y < frame.h; ++y)
	    {
	      for (x = 0; x < frame.w; ++x)
		{
		  i = y * frame.w + x;
		  ri = (y + 1) * frame.w - (x + 1);
		  frame.out[0][i] = frame.in[0][ri];
		}
	    }
	  for (y = 0; y < frame.Ch; ++y)
	    {
	      for (x = 0; x < frame.Cw; ++x)
		{
		  i = y * frame.Cw + x;
		  ri = (y + 1) * frame.Cw - (x + 1);
		  frame.out[1][i] = frame.in[1][ri];
		  frame.out[2][i] = frame.in[2][ri];
		}
	    }
	}
      if (flip_vertically)
	{
	  /* Flip the frame horizontally. */
	  for (y = 0; y < frame.h; ++y)

	    for (x = 0; x < frame.w; ++x)
	      {
		i = y * frame.w + x;
		ri = (frame.h - 1 - y) * frame.w + x;
		frame.out[0][i] = frame.in[0][ri];
	      }
	  for (y = 0; y < frame.Ch; ++y)
	    {
	      for (x = 0; x < frame.Cw; ++x)
		{
		  i = y * frame.Cw + x;
		  ri = (frame.Ch - 1 - y) * frame.Cw + x;
		  frame.out[1][i] = frame.in[1][ri];
		  frame.out[2][i] = frame.in[2][ri];
		}
	    }
	}

      if (verbose)
	mjpeg_log (LOG_INFO, "Writing Frame: %d", frame_number);
      frame_number++;

      /* Write the flipped frame. */
      errno = y4m_write_frame (1, &streaminfo, &frameinfo, frame.out);
    }

  /* did stream end unexpectedly ? */
  if (errno != Y4M_ERR_EOF && errno != Y4M_OK)
    mjpeg_error_exit1 ("%s", y4m_strerr (errno));

  /* Exit gently */
  return (0);
}
