/*
 *  yuvframestofields - halves the frame rate by taking one field per frame
 *
 *  Copyright (C) 2005, Jerome Cornet <jcornet@free.fr>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
#define HAVE_STDINT_H

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <mjpegtools/yuv4mpeg.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>

/* dominance =  0 for top field, 1 for bottom */
void
average_empty (unsigned char *yuv[3], y4m_stream_info_t * sinfo,
	       int dominance, unsigned char *interp[3])
{
  int i, j;
  int width, height;
  int index;
  int component;

  /* Interpolate each component */
  for (component = 0; component < 3; component++)
    {
      height = y4m_si_get_height (sinfo);
      width = y4m_si_get_width (sinfo);

      /* The chroma is usually subsampled */
      if (component)
	{
	  height =
	    (int) (double) height
	    * Y4M_RATIO_DBL (y4m_chroma_ss_y_ratio
			     (y4m_si_get_chroma (sinfo)));
	  width =
	    (int) (double) width
	    * Y4M_RATIO_DBL (y4m_chroma_ss_x_ratio
			     (y4m_si_get_chroma (sinfo)));
	}

      if (dominance == 0)
	{
	  for (i = 0; i < width; i++)
	    {
	      for (j = 0; j < (height / 2) - 1; j++)
		{
		  index = i + j * width;
		  interp[component][index] =
		    (yuv[component][index] +
		     yuv[component][index + width]) >> 1;
		}
	      index = i + j * width;
	      interp[component][index] = yuv[component][index];
	    }
	}
      else
	{
	  for (i = 0; i < width; i++)
	    {
	      interp[component][i] = yuv[component][i];

	      for (j = 1; j < (height / 2); j++)
		{
		  index = i + j * width;
		  interp[component][index] =
		    (yuv[component][index] +
		     yuv[component][index - width]) >> 1;
		}

	    }
	}
    }
}

int
main (int argc, char *argv[])
{

  int stream0 = STDIN_FILENO;	/* read from stdin */
  int outstream = STDOUT_FILENO;	/* output to stdout */
  int w, h;
  unsigned char *yuv_top[3], *yuv_bottom[3], *yuv_interp_t[3],
    *yuv_interp_b[3];
  y4m_stream_info_t sinfo0;
  y4m_stream_info_t osinfo;
  y4m_frame_info_t finfo0;
  int err;
  y4m_ratio_t framerate;
  char c;
  int verbose = 0;
  int dominance = Y4M_ILACE_NONE;	/* Y4M_ILACE_NONE Y4M_ILACE_TOP_FIRST Y4M_ILACE_BOTTOM_FIRST */
  int source_dominance;
  int framenumber = 0;
  int interpolate = 0;
  int component;

  while ((c = getopt (argc, argv, "hvtbi")) != -1)
    {
      switch (c)
	{
	case 'h':
	  {
	    mjpeg_log (LOG_INFO,
		       " -------------------------------------------------");
	    mjpeg_log (LOG_INFO,
		       "            Framerate doubler/deinterlacer            ");
	    mjpeg_log (LOG_INFO,
		       " -------------------------------------------------");

	    mjpeg_log (LOG_INFO, " Usage ");

	    mjpeg_log (LOG_INFO, " -h: this message");
	    mjpeg_log (LOG_INFO, " -v: verbose output");
	    mjpeg_log (LOG_INFO,
		       " -t: force top field first (default=autodetect)");
	    mjpeg_log (LOG_INFO,
		       " -b: force bottom field first (default=autodetect)");
	    mjpeg_log (LOG_INFO, " -i: average empty lines ");
	    exit (0);
	  }
	  break;
	case 'v':
	  verbose = 1;
	  break;
	case 't':
	  dominance = Y4M_ILACE_TOP_FIRST;
	  break;
	case 'b':
	  dominance = Y4M_ILACE_BOTTOM_FIRST;
	  break;
	case 'i':
	  interpolate = 1;
	  break;
	case '?':
	  mjpeg_error_exit1 ("Missing argument to option %c", c);
	  break;
	default:
	  mjpeg_error_exit1 ("Unknown Option %c", c);
	}
    }


  y4m_init_stream_info (&sinfo0);
  y4m_init_stream_info (&osinfo);
  y4m_init_frame_info (&finfo0);

  if ((err = y4m_read_stream_header (stream0, &sinfo0)) != Y4M_OK)
    mjpeg_error_exit1 ("Failed to input stream header:  %s errno=%d",
		       y4m_strerr (err), errno);

  if (y4m_si_get_plane_count (&sinfo0) != 3)
    mjpeg_error_exit1 ("Only 3-plane formats supported.");

  mjpeg_info ("Input stream parameters:");
  y4m_log_stream_info (LOG_INFO, "1> ", &sinfo0);

  w = y4m_si_get_width (&sinfo0);
  h = y4m_si_get_height (&sinfo0);

  source_dominance = y4m_si_get_interlace (&sinfo0);

  if (dominance)
    {
/* The user set a dominance... does it correspond to the source material */
/* if the source_dominance = 0 , it is not specified */
      if ((source_dominance) && (dominance != source_dominance))
	mjpeg_log (LOG_WARN, "Warning, source stream is %s, you said %s",
		   (source_dominance ==
		    Y4M_ILACE_BOTTOM_FIRST) ? "Bottom-field-first" :
		   "Top-field-first",
		   (dominance ==
		    Y4M_ILACE_BOTTOM_FIRST) ? "Bottom-field-first" :
		   "Top-field-first");

    }
  else
    {
/* The user did not set a dominance.. try to infer from the source */
      if (source_dominance)
	dominance = source_dominance;
      else
	mjpeg_error_exit1
	  ("Please specify top field first or bottom field first");
    }


  y4m_copy_stream_info (&osinfo, &sinfo0);
  framerate = y4m_si_get_framerate (&sinfo0);
  y4m_ratio_reduce (&framerate);
  y4m_si_set_interlace (&osinfo, Y4M_ILACE_NONE);

  /*  normalize stupid framerate */
  if ((framerate.n == 29969999) && (framerate.d == 100000))	/* I hate this */
    framerate = y4m_fps_NTSC;

  framerate.n *= 2;
  y4m_ratio_reduce (&framerate);
  y4m_si_set_framerate (&osinfo, framerate);

  y4m_log_stream_info (LOG_INFO, "1> ", &osinfo);

  /* Allocate for the worst case (4:4:4) */
  for (component = 0; component < 3; component++)
    {
      yuv_top[component] = (unsigned char *) malloc (w * h);
      yuv_bottom[component] = (unsigned char *) malloc (w * h);

      /* When we are not averaging the empty rows, they are simply
       * replicated, so we use the same pointer */
      if (interpolate)
	{
	  yuv_interp_t[component] = (unsigned char *) malloc (w * h);
	  yuv_interp_b[component] = (unsigned char *) malloc (w * h);
	}
      else
	{
	  yuv_interp_t[component] = yuv_top[component];
	  yuv_interp_b[component] = yuv_bottom[component];
	}
    }

  if ((err = y4m_write_stream_header (outstream, &osinfo)) != Y4M_OK)
    mjpeg_error_exit1 ("Failed to write output header:  %s",
		       y4m_strerr (err));

  /* Main loop, while no error happened process the frames 
   * if an error occurs during processing, it will exit the 
   * loop using a break instruction */

  while ((err =
	  y4m_read_fields (stream0, &sinfo0, &finfo0, yuv_top,
			   yuv_bottom)) == Y4M_OK)
    {
      if (verbose)
	mjpeg_log (LOG_INFO, "Input Frame #%d", framenumber++);

      if (interpolate)
	{
	  average_empty (yuv_top, &sinfo0, 0, yuv_interp_t);
	  average_empty (yuv_bottom, &sinfo0, 1, yuv_interp_b);
	}

      if (dominance == Y4M_ILACE_TOP_FIRST)
	{			/* top field first */
	  if ((err =
	       y4m_write_fields (outstream, &osinfo, &finfo0, yuv_top,
				 yuv_interp_t)) != Y4M_OK)
	    break;
	  if ((err =
	       y4m_write_fields (outstream, &osinfo, &finfo0, yuv_interp_b,
				 yuv_bottom)) != Y4M_OK)
	    break;
	}
      else
	{			/* bottom field first */
	  if ((err =
	       y4m_write_fields (outstream, &osinfo, &finfo0, yuv_interp_b,
				 yuv_bottom)) != Y4M_OK)
	    break;
	  if ((err =
	       y4m_write_fields (outstream, &osinfo, &finfo0, yuv_top,
				 yuv_interp_t)) != Y4M_OK)
	    break;

	}
    }

  if (err != Y4M_ERR_EOF)
    mjpeg_error_exit1 ("Some error somewhere:  %s", y4m_strerr (err));

  /* Free all the allocated frames */
  for (component = 0; component < 3; component++)
    {
      free (yuv_top[component]);
      free (yuv_bottom[component]);
      if (interpolate)
	{
	  free (yuv_interp_t[component]);
	  free (yuv_interp_b[component]);
	}

    }

  y4m_fini_stream_info (&sinfo0);
  y4m_fini_stream_info (&osinfo);
  y4m_fini_frame_info (&finfo0);

  return 0;
}
