/*
 *  yuvframestofields - halves the frame rate by taking one field per frame
 *
 *  Copyright (C) 2005, Jerome Cornet <jcornet@free.fr>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
#define HAVE_STDINT_H

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <mjpegtools/yuv4mpeg.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>


static void
usage (char *name)
{
  fprintf (stderr, "Usage: %s < input.yuv > output.yuv \n", name);
  exit (1);
}


int
main (int argc, char *argv[])
{

  int stream0 = STDIN_FILENO;	/* read from stdin */
  int outstream = STDOUT_FILENO;	/* output to stdout */
  int w, h;
  unsigned char *yuv[3];
  y4m_stream_info_t sinfo0;
  y4m_stream_info_t osinfo;
  y4m_frame_info_t finfo0;
  int err;
  unsigned long ntsc_frame_counter = 0;
  unsigned long pal_field_counter = 0;
  unsigned long expected_pal_fields = 0;
  y4m_ratio_t framerate;



  if (argc > 1)
    usage (argv[0]);

  y4m_init_stream_info (&sinfo0);
  y4m_init_stream_info (&osinfo);
  y4m_init_frame_info (&finfo0);


  if ((err = y4m_read_stream_header (stream0, &sinfo0)) != Y4M_OK)
    mjpeg_error_exit1 ("Failed to input stream header:  %s errno=%d",
		       y4m_strerr (err), errno);

  mjpeg_info ("Input stream parameters:");
  y4m_log_stream_info (LOG_INFO, "1> ", &sinfo0);

  w = y4m_si_get_width (&sinfo0);
  h = y4m_si_get_height (&sinfo0);

  y4m_copy_stream_info (&osinfo, &sinfo0);
  framerate = y4m_si_get_framerate (&sinfo0);
  y4m_ratio_reduce (&framerate);
  y4m_si_set_interlace (&osinfo, Y4M_ILACE_NONE);

  if (Y4M_RATIO_EQL (framerate, y4m_fps_NTSC))
    y4m_si_set_framerate (&osinfo, y4m_fps_PAL_FIELD);
  if ((framerate.n == 29969999) && (framerate.d == 100000))	/* I hate this */
    y4m_si_set_framerate (&osinfo, y4m_fps_PAL_FIELD);
  else
    mjpeg_error_exit1 ("Input not NTSC video rate");

  y4m_log_stream_info (LOG_INFO, "1> ", &osinfo);


  yuv[0] = (unsigned char *) malloc (w * h);
  yuv[1] = (unsigned char *) malloc (w * h / 4);
  yuv[2] = (unsigned char *) malloc (w * h / 4);

  if ((err = y4m_write_stream_header (outstream, &osinfo)) != Y4M_OK)
    mjpeg_error_exit1 ("Failed to write output header:  %s",
		       y4m_strerr (err));

  while (err == Y4M_OK)
    {
/*  To transform NTSC frames to PAL fields, we replicate frames as such:
 *  NTSC:         30000 frames / 1001 seconds
 *  PAL_FIELD:    50000 fields / 1000 seconds == 50050 fields / 1001 seconds
 *  Therefore for every 600 NTSC frames we should have 1001 PAL fields
 *  (divide by 50)
 *  We use simple integer division to decide on whether to replicate the
 *  current frame. Every 600th frame will be replicated and every 1st will
 *  not so we can reset the counter after 600 frames.
 */

      err = y4m_read_frame (stream0, &sinfo0, &finfo0, yuv);
      if (err != Y4M_OK)
	break;

      expected_pal_fields = ++ntsc_frame_counter * 1001 / 600;
      while (pal_field_counter < expected_pal_fields) {
      	pal_field_counter++;
      	err = y4m_write_frame (outstream, &osinfo, &finfo0, yuv);
      }
      if (err != Y4M_OK)
      	break;
      if (ntsc_frame_counter == 600)
      	ntsc_frame_counter = pal_field_counter = 0;

      /* at this stage, if there was an error, err has it 
       * if you want to add more crap, add this here:
       * if (err != Y4M_OK) break; 
       */

    }
  if (err != Y4M_ERR_EOF)
    mjpeg_error_exit1 ("Some error somewhere:  %s", y4m_strerr (err));

  free (yuv[0]);
  free (yuv[1]);
  free (yuv[2]);

  y4m_fini_stream_info (&sinfo0);
  y4m_fini_stream_info (&osinfo);
  y4m_fini_frame_info (&finfo0);
  return 0;
}
